import logging
import os
import click
import yaml
from dame import _schema, _utils
#from contextlib import contextmanager

logger = logging.getLogger()
schemaDefinitions = _schema.get_schema()["definitions"]
error_log = ""


def make_yaml(download_path=None):

    # sets path, this determines where the yaml will be made. Default is the current directory
    if download_path is None:
        path = os.getcwd()
    else:
        path = download_path
    path = os.path.join(path, "wings-component-outline.yaml")

    if os.path.isfile(path):
        click.echo("\"" + path + "\" already exists. Do you want to overwrite it? [y/n]")
        ans = input()
        if not(ans.lower() == 'y' or ans.lower() == "yes"):
            logger.info("Aborting YAML Generation")
            exit(0)

    # Checks if directory exists and finds file
    try:
        stream = open(path, 'w+')
    except FileNotFoundError:
        click.echo("\"" + os.path.join(os.getcwd(), download_path) + "\" doesnt exists. Do you want to make it? [y/n]")
        ans = input()
        if ans.lower() == 'y' or ans.lower() == "yes":
            os.mkdir(download_path)
            stream = open(path, 'w+')
        else:
            exit(0)

    yaml_outline = write_properties(_schema.get_schema()["properties"])

    yaml.dump(yaml_outline, stream, sort_keys=False)

def make_yaml_from_dict(download_path=None, dict=None):

    # sets path, this determines where the yaml will be made. Default is the current directory
    if download_path is None:
        path = os.getcwd()
    else:
        path = download_path
    path = os.path.join(path, "setup_name.yaml")

    if os.path.isfile(path):
        click.echo("\"" + path + "\" already exists. Do you want to overwrite it? [y/n]")
        ans = input()
        if not(ans.lower() == 'y' or ans.lower() == "yes"):
            logger.info("Aborting YAML Generation")
            exit(0)

    # Checks if directory exists and finds file
    try:
        stream = open(path, 'w+')
    except FileNotFoundError:
        click.echo("\"" + os.path.join(os.getcwd(), download_path) + "\" doesnt exists. Do you want to make it? [y/n]")
        ans = input()
        if ans.lower() == 'y' or ans.lower() == "yes":
            os.mkdir(download_path)
            stream = open(path, 'w+')
        else:
            exit(0)

    yaml.dump(dict, stream, sort_keys=False)

def write_properties(prop):
    dict = {}

    for i in prop:
        # print(i + ": " + str(type((prop[i])["type"])))
        curr = prop[i]

        try:
            # if there is only one type for the current property
            if type(curr["type"]) is str:
                if curr["type"] == "string":
                    if i == "schemaVersion":
                        dict[i] = _schema.get_schema_version()
                    else:
                        dict[i] = ""
                elif curr["type"] == "array":
                    ci = curr["items"]
                    if "$ref" in list(ci.keys()):
                        ref = ci["$ref"]
                        ref = ref.split('/')
                        ref = ref[-1]
                        dict[i] = [write_properties((schemaDefinitions[ref])["properties"])]
                    elif "type" in list(ci.keys()):
                        dict[i] = []
                    else:
                        logger.warning("unknown type fount in " + i)

                elif curr["type"] == "integer":
                    dict[i] = 0
                elif curr["type"] == "float":
                    dict[i] = 0.0
                elif curr["type"] == "boolean":
                    dict[i] = False
                elif curr["type"] == "object":
                    try:
                        dict[i] = write_properties(curr["properties"])
                    except KeyError as ke:
                        logger.info("In object: \"" + i + "\" couldn't find type: " + str(ke) + "making empty object")
                        dict[i] = {}
            # if there are multiple types for the given property. Favoring in order: objects, arrays, everything else
            elif type(curr["type"]) is list:
                if "object" in curr["type"]:
                    try:
                        dict[i] = write_properties(curr["properties"])
                    except KeyError as ke:
                        logger.info("In object: \"" + i + "\" couldn't find type " + str(ke) + " making empty object")
                        dict[i] = {}
                elif "array" in curr["type"]:
                    dict[i] = []
                else:
                    dict[i] = ""

        except KeyError as err:
            logger.warning("\"KeyError\" in \"" + i + "\" no \"" + str(err) + "\"")

    return dict


def _main():
    make_yaml()


if __name__ == "__main__":
    try:
        _main()
    except Exception as e:
        logger.exception(e)



# -*- coding: utf-8 -*-
"""
dame.

:license: Apache 2.0
"""
import json
import logging
import os
import re
from pathlib import Path

import click

import semver
import dame
from dame.cli_methods import run_method, edit_inputs_model_configuration, edit_parameter_config_or_setup, \
    verify_input_parameters, run_method_setup
from dame.downloader import check_size, parse_inputs, parse_outputs
from dame.emulatorapi import get_summary, list_summaries, obtain_results
from dame.utils import obtain_id, download_file, download_data_file, humansize, SERVER, check_is_none
from dame.modelcatalogapi import get_setup, list_setup, get_model, list_model_configuration, get_model_configuration
from dame import _utils, _makeyaml
from dame._utils import log
import texttable as tt
from modelcatalog import DatasetSpecification, SampleResource, ApiValueError, OpenApiException

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


@click.group()
@click.option("--verbose", "-v", default=0, count=True)
def cli(verbose):
    _utils.init_logger()
    lv = ".".join(_utils.get_latest_version().split(".")[:3])
    cv = ".".join(dame.__version__.split(".")[:3])

    if semver.compare(lv, cv) > 0:
        click.secho(
            f"""WARNING: You are using dame version {dame.__version__}, however version {lv} is available.
You should consider upgrading via the 'pip install --upgrade dame' command.""",
            fg="yellow",
        )


@cli.command(help="Open the Model Catalog in your browser")
def browse():
    click.launch('https://models.mint.isi.edu')

"""
Run a modelconfiguration or modelconfiguration
"""
@cli.command(help="Run a model configuration or model configuration setup")
@click.argument(
    "name",
    type=click.STRING
)
def run(name):
    try:
        config = get_model_configuration(name)
    except OpenApiException as e:
        logging.error(e.reason)
        exit(0)

    if "ModelConfigurationSetup" in config.type:
        resource = get_setup(name)
        verify_input_parameters(resource)
    elif "ModelConfiguration" in config.type:
        resource = get_model_configuration(name)
        verify_input_parameters(resource)
    # setup = get_setup(name)
    # edit_inputs_setup(setup)
    run_method_setup(resource)



# -*- coding: utf-8 -*-
import csv
import json
import logging.config
import os
import requests
import importlib
log = logging.getLogger()

def load_module(module_name):
    try:
        return importlib.import_module(module_name)
    except ModuleNotFoundError:
        log.error("Model <%s> not found" % module_name)
        raise

def init_logger():
    LOGLEVEL = logging.DEBUG
    logging.config.dictConfig({
        'version': 1    ,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                # exact format is not important, this is the minimum information
                'format': '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
            },
        },
        'handlers': {
            # console logs to stderr
            'console': {
                'class': 'logging.StreamHandler',
                'formatter': 'default',
            },
        },
        'loggers': {
            # default for all undefined Python modules
            '': {
                'level': 'DEBUG',
                'handlers': ['console'],
            },
            # Our application code
            'wcm': {
                'level': LOGLEVEL,
                'handlers': ['console'],
                # Avoid double logging because of root logger
                'propagate': False,
            },
            # Prevent noisy modules from logging to Sentry
            'noisy_module': {
                'level': 'ERROR',
                'handlers': ['console'],
                'propagate': False,
            },
        },
    })
    logger = logging.getLogger(__package__)

def get_latest_version():
    return requests.get("https://pypi.org/pypi/dame-cli/json").json()["info"]["version"]



import itertools
import logging
import os
import uuid
from pathlib import Path

import click
import texttable as tt
from dame.utils import check_is_none, create_yaml_from_resource, obtain_id
from dame.executor import execute_setup
from dame._utils import log
from dame.modelcatalogapi import get_setup
from modelcatalog import ApiException, SampleResource

def verify_input_parameters(model_configuration):
    for _input in model_configuration.has_input:
        if not hasattr(_input, "has_fixed_resource"):
            click.secho("The information of the setup is incomplete", fg="yellow")
            print_data_property_table(_input)
            url = click.prompt('Please, enter the url of the previous input', type=click.STRING)
            s = SampleResource(id="https://w3id.org/okn/i/mint/".format(str(uuid.uuid4())),
                               data_catalog_identifier="FFF-3s5c112e-c7ae-4cda-ba23-2e4f2286a18o",
                               value=[url])
            _input.has_fixed_resource = [s.to_dict()]
    click.secho("The information of the setup is complete", fg="green")
    return model_configuration


def edit_parameter_config_or_setup(resource, auto=False):
    """not used"""
    for parameter in resource.has_parameter:
        logging.info("Checking {}".format(parameter))
        logging.info("Checking {}".format(check_is_none(parameter, 'id')))
        _id = obtain_id(check_is_none(parameter, 'id'))
        default_value = check_is_none(parameter, 'has_default_value')
        print_data_property_table(parameter)
        if not default_value:
            value = click.prompt('Enter the value for the parameter.')
        else:
            default_value = default_value[0]
            if auto:
                value = default_value
                click.echo("Using the default valuer {}".format(default_value))
            else:
                value = click.prompt('Enter the value for the parameter:', default=default_value)
        parameter["hasFixedValue"] = [value]

def print_data_property_table(resource):
    resource_dict = resource.to_dict()
    tab = tt.Texttable(max_width=100)
    headings = ['Property', 'Value']
    tab.header(headings)
    for key, value in resource_dict.items():
        if isinstance(value, dict) or key == "type" or key == "has_presentation":
            continue
        tab.add_row([key,value])
    print(tab.draw())


def edit_parameter_config_setup(resource):
    print_table(resource.to_dict())


def edit_inputs_model_configuration(model_configuration):
    for _input in model_configuration.has_input:
        print("=======================================================")
        _id = check_is_none(_input, 'id')
        description = check_is_none(_input, 'description')
        label = check_is_none(_input, 'label')
        _format = check_is_none(_input, 'format')
        print_data_property_table(_input)
        url = click.prompt('Please enter the url of the input shown above', type=click.STRING)
        s = SampleResource(data_catalog_identifier="FFF-3s5c112e-c7ae-4cda-ba23-2e4f2286a18o",
                           value=url,
                           description=description,
                           label=label)
        _input.has_fixed_resource = [s]

def edit_setup(setup):
    """
    Edit the inputs and parameters of a setup
    """
    edit_parameter_config_setup(resource=setup)


def run_method(editable, name):
    """
    Call download_setup(): Download the setup(s) as yaml file
    Call execute_setup(): Read the yaml file and execute
    """
    click.secho("Downloading the inputs and parameters", fg="green")
    setup = get_setup(name)
    name = obtain_id(setup.id)
    if editable:
        edit_setup(setup)

    try:
        file_path = create_yaml_from_resource(resource=setup, name=name, output=Path('.'))
        read_and_execute(file_path)
    except ApiException as e:
        click.secho("Unable to download the setup {}".format(e), fg="red")
        exit(1)


def run_method_setup(setup):
    """
    Call download_setup(): Download the setup(s) as yaml file
    Call execute_setup(): Read the yaml file and execute
    """
    name = obtain_id(setup.id)
    try:
        file_path = create_yaml_from_resource(resource=setup, name=name, output=Path('.'))
        read_and_execute(file_path)
    except ApiException as e:
        click.secho("Unable to download the setup {}".format(e), fg="red")
        exit(1)


def read_and_execute(file_path):
    click.secho("Executing the setup", fg="green")
    status, file_dir = execute_setups(file_path)
    for setup in status:
        if setup["exitcode"] == 0:
            click.secho("[{}] The execution has been successful".format(setup["name"]), fg="green")
            click.secho("[{}] Results available at: {} ".format(setup["name"], file_dir), fg="green")
        else:
            click.secho("[{}] The execution has failed".format(setup["name"]), fg="red")


def execute_setups(path):
    """
    Find the setup files if the path is a directory and execute it
    """
    setup_files = find_setup_files(path)
    try:
        status, file_dir = execute_setup(setup_files)
    except Exception as e:
        log.error(e, exc_info=True)
        exit(1)
    return status, file_dir


def find_setup_files(path):
    setup_files = []
    if os.path.isdir(path):
        path = Path(path)
        for file in os.listdir(path):
            if file.endswith(".yaml") or file.endswith(".yml"):
                setup_files.append(path / file)
    elif os.path.isfile(path):
        default_path = Path('.')
        setup_files.append(default_path / path)
    return setup_files



from __future__ import print_function
import modelcatalog
from dame._utils import log
from modelcatalog.rest import ApiException

configuration = modelcatalog.Configuration()
USERNAME = "mint@isi.edu"
custom_query_name_config = ""

def get_model(_id):
    api_instance = modelcatalog.ModelApi()
    try:
        api_response = api_instance.models_id_get(id=_id)
        return api_response
    except ApiException as e:
        raise e

def list_model_configuration(label=None):
   api_instance = modelcatalog.ModelConfigurationApi()
   try:
       api_response = api_instance.modelconfigurations_get(username=USERNAME)
       return api_response
   except ApiException as e:
       raise e


def get_model_configuration(_id):
    api_instance = modelcatalog.ModelConfigurationApi()
    try:
        api_response = api_instance.custom_modelconfigurations_id_get(_id, username=USERNAME)
        return api_response
    except ApiException as e:
        raise e


def get_setup(_id):
    # create an instance of the API class
    api_instance = modelcatalog.ModelConfigurationSetupApi()
    try:
        # Get a ModelConfiguration
        api_response = api_instance.custom_modelconfigurationsetups_id_get(_id, username=USERNAME)
        return api_response
    except ApiException as e:
        raise e


def list_setup(label=None):
    # create an instance of the API class
    api_instance = modelcatalog.ConfigurationSetupApi()
    try:
        # Get a ModelConfigurationSetup
        api_response = api_instance.configurationsetups_get(username=USERNAME)
        return api_response
    except ApiException as e:
        raise e

def datasetspecifications_id_get(_id):
    try:
        # Get a DatasetSpecification
        api_instance = modelcatalog.DatasetSpecificationApi()
        api_response = api_instance.datasetspecifications_id_get(_id, username=USERNAME)
        return api_response
    except ApiException as e:
        print("Exception when calling DatasetSpecificationApi->datasetspecifications_id_get: %s\n" % e)



import pathlib
import requests
import os
import tempfile
from zipfile import ZipFile
import validators
import yaml
from dame.modelcatalogapi import get_setup, datasetspecifications_id_get

ignore_dirs = ["__MACOSX"]
SERVER = "https://dev.mint.isi.edu"

def convert_object_to_dict(o):
    if isinstance(o, object):
        return o.to_dict()
    return o

def check_is_none(item, key):
     item[key] if key in item else ''


def create_yaml_from_resource(resource, name, output):
    filename = name + ".yaml"
    path = pathlib.Path.cwd() / output / filename
    with open(path, mode='w+') as fid:
        yaml.dump(resource, fid)
    return path


def obtain_id(url):
    if validators.url(url):
        return url.split('/')[-1]
    return url

def download_extract_zip(url, _dir, setup_name):
    temp = tempfile.NamedTemporaryFile(prefix="component_")
    content = download_file(url)
    temp.write(content)
    with ZipFile(temp.name, 'r') as zip:
        zip.extractall(_dir)
    directories = os.listdir(_dir)
    if isinstance(directories, list):
        try:
            for ignore_dir in ignore_dirs:
                directories.remove(ignore_dir)
        except:
            pass

    if len(directories) == 1:
        return os.path.join(_dir, directories[0])
    else:
        raise ValueError("The zipfile must has one directory.")


def download_file(url):
    headers={'Cache-Control': 'no-cache'}
    r = requests.get(url, allow_redirects=True, headers=headers)
    try:
        r.raise_for_status()
    except requests.exceptions.HTTPError:
        raise requests.exceptions.HTTPError(r)
    except requests.exceptions.RequestException:
        raise requests.exceptions.RequestException(r)
    return r.content

def validate_suffix(suffix):
    return suffix if suffix.startswith('.') else ".{}".format(suffix)

def download_data_file(url, _dir, data_specification_id=None):
    headers={'Cache-Control': 'no-cache'}
    r = requests.get(url, allow_redirects=True, headers=headers)
    filename = url.split('/')[-1]
    filepath = pathlib.Path(_dir / filename)
    if filepath.suffix == "" and data_specification_id:
        file_extension = datasetspecifications_id_get(obtain_id(data_specification_id)).has_format[0]
        if file_extension:
            filepath = filepath.with_suffix(validate_suffix(file_extension))
    with requests.get(url, stream=True, headers=headers) as r:
        r.raise_for_status()
        with open(filepath, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk: # filter out keep-alive new chunks
                    f.write(chunk)
    return filepath, filename


suffixes = ['B', 'KB', 'MB', 'GB', 'TB', 'PB']
def humansize(nbytes):
    i = 0
    while nbytes >= 1024 and i < len(suffixes)-1:
        nbytes /= 1024.
        i += 1
    f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
    return '%s %s' % (f, suffixes[i])



import logging
import platform
import subprocess
import uuid
from pathlib import Path

from yaml import load, Loader

from dame.utils import download_data_file, download_extract_zip, obtain_id, convert_object_to_dict
from dame._utils import log

KEYS_REQUIRED_PARAMETER = {"has_default_value", "position"}
KEYS_REQUIRED_OUTPUT = {"label", "has_format", "position"}
KEYS_REQUIRED_INPUT = {"has_fixed_resource"}

if platform.system() == "Linux":
    SINGULARITY_CWD_LINE = "/usr/bin/singularity exec docker://{} ./run"
elif platform.system() == "Darwin":
    SINGULARITY_CWD_LINE = "/usr/local/bin/singularity exec docker://{} ./run"

EXECUTION_DIRECTORY = "executions"

def build_input(inputs, _dir):
    line = ""
    for _input in inputs:
        _input = convert_object_to_dict(_input)
        if not _input.keys() >= KEYS_REQUIRED_INPUT:
            raise ValueError(f'{_input["id"]} has not a fixedResource')
        url = _input["has_fixed_resource"][0]["value"][0]
        file_path, file_name = download_data_file(url, _dir)
        position = _input["position"][0]
        line += " -i{} {}".format(position, file_name)
    return line


def build_output(outputs):
    line = ""
    for _output in outputs:
        _output = convert_object_to_dict(_output)
        if not _output.keys() >= KEYS_REQUIRED_OUTPUT:
            raise ValueError(f'{_output["id"]}  has not the required information')
        label = _output["label"][0]
        extension = _output["has_format"][0]
        position = _output["position"][0]
        line += " -o{} {}.{}".format(position, label, extension)
    return line


def build_parameter(parameters):
    line = ""
    for _parameter in parameters:
        _parameter = convert_object_to_dict(_parameter)
        if not _parameter.keys() >= KEYS_REQUIRED_PARAMETER:
            raise ValueError(f'{_parameter["id"]} has not the required information ')
        if "has_fixed_resource" in _parameter:
            value = _parameter["has_fixed_resource"][0]
        else:
            value = _parameter["has_default_value"][0]
        position = _parameter["position"][0]
        line += " -p{} {}".format(position, value)
    return line


def build_command_line(resource, _dir):
    setup_name = obtain_id(resource.id)
    inputs = resource.has_input
    try:
        parameters = resource.has_parameter
    except:
        parameters = None
    outputs = resource.has_output
    component_url = resource.has_component_location[0]
    software_image = resource.has_software_image
    if software_image:
        has_software_image = software_image[0].label[0]
        line = SINGULARITY_CWD_LINE.format(has_software_image)
    else:
        raise ValueError("Software is not available")
    component_dir = download_extract_zip(component_url, _dir, setup_name)
    path = Path(component_dir)
    src_path = path / "src"
    if inputs:
        l = build_input(inputs, src_path)
        line += " {}".format(l)
    if outputs:
        l = build_output(outputs)
        line += " {}".format(l)
    if parameters is not None:
        l = build_parameter(parameters)
        line += " {}".format(l)
    return line, src_path


def execute_setup(setup_paths):
    data = []
    _dir = Path("%s/" % EXECUTION_DIRECTORY)
    _dir.mkdir(parents=True, exist_ok=True)

    for setup in setup_paths:
        setup_dict = load((setup).open(), Loader=Loader)
        setup_name = obtain_id(setup_dict.id)
        execution_dir = "{}/{}_{}".format(_dir, setup_name, uuid.uuid1())
        execution_dir_path = Path(execution_dir)
        execution_dir_path.mkdir(parents=True, exist_ok=True)
        try:
            setup_cmd_line, cwd_path = build_command_line(setup_dict, execution_dir_path)
        except Exception as e:
            raise e
        log_file_path = "{}/output.log".format(execution_dir)
        data.append({
            "cmd": setup_cmd_line.split(' '),
            "log": open(log_file_path, 'wb'),
            "directory": cwd_path,
            "name": setup_name
        })
        log.info(f'Execution {setup_name} running,  check the logs on {log_file_path}')

    _ = [subprocess.Popen(["chmod", "+x", "run"], stdout=item["log"], stderr=item["log"], cwd=item["directory"]) for item
        in data]
    procs_list = [subprocess.Popen(item["cmd"], stdout=item["log"], stderr=item["log"], cwd=item["directory"]) for item
                  in data]

    for proc in procs_list:
        proc.wait()

    status = []
    for proc, item in zip(procs_list, data):
        status.append({
            "exitcode": proc.returncode,
            "name": item["name"]
        })

    return status, cwd_path



from __future__ import print_function
import ingestion
from ingestion import Thread
from dame.downloader import get_json
from ingestion.rest import ApiException

def get_summary_from_dict(dict):
    thread_dict = dict['thread']
    thread = Thread(id=thread_dict["id"],
                    text=thread_dict["text"],
                    time_period=thread_dict["time_period"],
                    indicators=thread_dict["indicators"],
                    models=thread_dict["models"],
                    datasets=thread_dict["datasets"],
                    results=thread_dict["results"]
                    )
    return thread

def get_summary(thread_id):
    api_instance = ingestion.SummaryApi()
    try:
        api_response = api_instance.summary_thread_id_get(thread_id)
        return api_response
    except ApiException as e:
        raise e


def list_summaries(limit=200, page=1, model=""):
    # create an instance of the API class
    api_instance = ingestion.SummaryApi()
    try:
        api_response = api_instance.summary_get(limit=limit, page=page, model=model)
        print(api_response)
        return api_response
    except ApiException as e:
        raise e

def obtain_results(thread_id, force=True):
    api_instance = ingestion.ResultApi()
    try:
        # Get a result
        print(thread_id)
        api_response = api_instance.results_thread_id_get(thread_id, force)
        return api_response
    except ApiException as e:
        raise ValueError("Not found")


def create_csv(scenario_id, problem_id, thread_id):
    api_instance = ingestion.SummaryApi()
    modelthread = ingestion.Modelthread(scenario_id=scenario_id, subgoal_id=problem_id, thread_id=thread_id)
    try:
        # Create a summary
        api_response = api_instance.summary_post(modelthread)
        return api_response
    except ApiException as e:
        print("Exception when calling DefaultApi->get_summary: %s\n" % e)
